__version__ = "0.6.1"

from .const import (
    ATTR_ACCELERATION,
    ATTR_ALARM,
    ATTR_BATTERY,
    ATTR_CARBON_MONOXIDE,
    ATTR_CODE_CHANGED,
    ATTR_CODE_LENGTH,
    ATTR_COLOR_MODE,
    ATTR_COLOR_NAME,
    ATTR_COLOR_TEMP,
    ATTR_CONTACT,
    ATTR_CURRENT,
    ATTR_DEVICE_ID,
    ATTR_DOOR,
    ATTR_DOUBLE_TAPPED,
    ATTR_ENERGY,
    ATTR_ENERGY_SOURCE,
    ATTR_ENTRY_DELAY,
    ATTR_EXIT_DELAY,
    ATTR_HELD,
    ATTR_HUE,
    ATTR_HUMIDITY,
    ATTR_ILLUMINANCE,
    ATTR_LAST_CODE_NAME,
    ATTR_LEVEL,
    ATTR_LOCK,
    ATTR_LOCK_CODES,
    ATTR_MAX_CODES,
    ATTR_MOTION,
    ATTR_NAME,
    ATTR_NUM_BUTTONS,
    ATTR_POSITION,
    ATTR_POWER,
    ATTR_POWER_SOURCE,
    ATTR_PRESENCE,
    ATTR_PRESSURE,
    ATTR_PUSHED,
    ATTR_SATURATION,
    ATTR_SECURITY_KEYPAD,
    ATTR_SMOKE,
    ATTR_SPEED,
    ATTR_SWITCH,
    ATTR_TEMPERATURE,
    ATTR_UV,
    ATTR_VALUE,
    ATTR_VOLTAGE,
    ATTR_WATER,
    ATTR_WINDOW_SHADE,
    CAP_ALARM,
    CAP_COLOR_CONTROL,
    CAP_COLOR_MODE,
    CAP_COLOR_TEMP,
    CAP_CONTACT_SENSOR,
    CAP_DOOR_CONTROL,
    CAP_DOUBLE_TAPABLE_BUTTON,
    CAP_ENERGY_METER,
    CAP_ENERGY_SOURCE,
    CAP_FAN_CONTROL,
    CAP_GARAGE_DOOR_CONTROL,
    CAP_HOLDABLE_BUTTON,
    CAP_ILLUMINANCE_MEASUREMENT,
    CAP_LIGHT,
    CAP_LOCK,
    CAP_LOCK_CODES,
    CAP_MOTION_SENSOR,
    CAP_MUSIC_PLAYER,
    CAP_POWER_METER,
    CAP_POWER_SOURCE,
    CAP_PRESENCE_SENSOR,
    CAP_PRESSURE_MEASUREMENT,
    CAP_PUSHABLE_BUTTON,
    CAP_RELATIVE_HUMIDITY_MEASUREMENT,
    CAP_SECURITY_KEYPAD,
    CAP_SWITCH,
    CAP_SWITCH_LEVEL,
    CAP_TEMPERATURE_MEASUREMENT,
    CAP_THERMOSTAT,
    CAP_WINDOW_SHADE,
    CMD_ARM_AWAY,
    CMD_ARM_HOME,
    CMD_ARM_NIGHT,
    CMD_AUTO,
    CMD_AWAY,
    CMD_BOTH,
    CMD_CLOSE,
    CMD_COOL,
    CMD_CYCLE_SPEED,
    CMD_DELETE_CODE,
    CMD_DISARM,
    CMD_ECO,
    CMD_EMERGENCY_HEAT,
    CMD_FAN_AUTO,
    CMD_FAN_CIRCULATE,
    CMD_FAN_ON,
    CMD_FLASH,
    CMD_GET_CODES,
    CMD_HEAT,
    CMD_LOCK,
    CMD_OFF,
    CMD_ON,
    CMD_OPEN,
    CMD_PRESENT,
    CMD_SET_CODE,
    CMD_SET_CODE_LENGTH,
    CMD_SET_COLOR,
    CMD_SET_COLOR_TEMP,
    CMD_SET_COOLING_SETPOINT,
    CMD_SET_ENTRY_DELAY,
    CMD_SET_EXIT_DELAY,
    CMD_SET_FAN_MODE,
    CMD_SET_HEATING_SETPOINT,
    CMD_SET_HUE,
    CMD_SET_LEVEL,
    CMD_SET_POSITION,
    CMD_SET_PRESENCE,
    CMD_SET_SAT,
    CMD_SET_SPEED,
    CMD_SET_THERMOSTAT_MODE,
    CMD_SIREN,
    CMD_STROBE,
    CMD_UNLOCK,
    COLOR_MODE_CT,
    COLOR_MODE_RGB,
    DEFAULT_FAN_SPEEDS,
    HSM_ARM_ALL,
    HSM_ARM_AWAY,
    HSM_ARM_HOME,
    HSM_ARM_NIGHT,
    HSM_ARM_RULES,
    HSM_CANCEL_ALERTS,
    HSM_DISARM,
    HSM_DISARM_ALL,
    HSM_DISARM_RULES,
    HSM_STATUS_ALL_DISARMED,
    HSM_STATUS_ARMED_AWAY,
    HSM_STATUS_ARMED_HOME,
    HSM_STATUS_ARMED_NIGHT,
    HSM_STATUS_ARMING_AWAY,
    HSM_STATUS_ARMING_HOME,
    HSM_STATUS_ARMING_NIGHT,
    HSM_STATUS_DISARMED,
    ID_HSM_STATUS,
    ID_MODE,
    STATE_ARMED_AWAY,
    STATE_ARMED_HOME,
    STATE_ARMED_NIGHT,
    STATE_CLOSED,
    STATE_CLOSING,
    STATE_DISARMED,
    STATE_LOCKED,
    STATE_LOW,
    STATE_OFF,
    STATE_ON,
    STATE_OPEN,
    STATE_OPENING,
    STATE_PARTIALLY_OPEN,
    STATE_UNKNOWN,
    STATE_UNLOCKED,
    STATE_UNLOCKED_WITH_TIMEOUT,
)
from .error import ConnectionError, InvalidConfig, InvalidToken, RequestError
from .hub import Hub
from .types import Attribute, Device, Event

__all__ = [
    "ATTR_ACCELERATION",
    "ATTR_ALARM",
    "ATTR_BATTERY",
    "ATTR_CARBON_MONOXIDE",
    "ATTR_CODE_CHANGED",
    "ATTR_CODE_LENGTH",
    "ATTR_COLOR_MODE",
    "ATTR_COLOR_NAME",
    "ATTR_COLOR_TEMP",
    "ATTR_CONTACT",
    "ATTR_CURRENT",
    "ATTR_DEVICE_ID",
    "ATTR_DOOR",
    "ATTR_DOUBLE_TAPPED",
    "ATTR_ENERGY",
    "ATTR_ENERGY_SOURCE",
    "ATTR_ENTRY_DELAY",
    "ATTR_EXIT_DELAY",
    "ATTR_HELD",
    "ATTR_HUE",
    "ATTR_HUMIDITY",
    "ATTR_ILLUMINANCE",
    "ATTR_LAST_CODE_NAME",
    "ATTR_LEVEL",
    "ATTR_LOCK",
    "ATTR_LOCK_CODES",
    "ATTR_MAX_CODES",
    "ATTR_MOTION",
    "ATTR_NAME",
    "ATTR_NUM_BUTTONS",
    "ATTR_POSITION",
    "ATTR_POWER",
    "ATTR_POWER_SOURCE",
    "ATTR_PRESENCE",
    "ATTR_PRESSURE",
    "ATTR_PUSHED",
    "ATTR_SATURATION",
    "ATTR_SECURITY_KEYPAD",
    "ATTR_SMOKE",
    "ATTR_SPEED",
    "ATTR_SWITCH",
    "ATTR_TEMPERATURE",
    "ATTR_UV",
    "ATTR_VALUE",
    "ATTR_VOLTAGE",
    "ATTR_WATER",
    "ATTR_WINDOW_SHADE",
    "Attribute",
    "CAP_ALARM",
    "CAP_COLOR_CONTROL",
    "CAP_COLOR_MODE",
    "CAP_COLOR_TEMP",
    "CAP_CONTACT_SENSOR",
    "CAP_DOOR_CONTROL",
    "CAP_DOUBLE_TAPABLE_BUTTON",
    "CAP_ENERGY_METER",
    "CAP_ENERGY_SOURCE",
    "CAP_FAN_CONTROL",
    "CAP_GARAGE_DOOR_CONTROL",
    "CAP_HOLDABLE_BUTTON",
    "CAP_ILLUMINANCE_MEASUREMENT",
    "CAP_LIGHT",
    "CAP_LOCK",
    "CAP_LOCK_CODES",
    "CAP_MOTION_SENSOR",
    "CAP_MUSIC_PLAYER",
    "CAP_POWER_METER",
    "CAP_POWER_SOURCE",
    "CAP_PRESENCE_SENSOR",
    "CAP_PRESSURE_MEASUREMENT",
    "CAP_PUSHABLE_BUTTON",
    "CAP_RELATIVE_HUMIDITY_MEASUREMENT",
    "CAP_SECURITY_KEYPAD",
    "CAP_SWITCH",
    "CAP_SWITCH_LEVEL",
    "CAP_TEMPERATURE_MEASUREMENT",
    "CAP_THERMOSTAT",
    "CAP_WINDOW_SHADE",
    "CMD_ARM_AWAY",
    "CMD_ARM_HOME",
    "CMD_ARM_NIGHT",
    "CMD_AUTO",
    "CMD_AWAY",
    "CMD_BOTH",
    "CMD_CLOSE",
    "CMD_COOL",
    "CMD_CYCLE_SPEED",
    "CMD_DELETE_CODE",
    "CMD_DISARM",
    "CMD_ECO",
    "CMD_EMERGENCY_HEAT",
    "CMD_FAN_AUTO",
    "CMD_FAN_CIRCULATE",
    "CMD_FAN_ON",
    "CMD_FLASH",
    "CMD_GET_CODES",
    "CMD_HEAT",
    "CMD_LOCK",
    "CMD_OFF",
    "CMD_ON",
    "CMD_OPEN",
    "CMD_PRESENT",
    "CMD_SET_CODE",
    "CMD_SET_CODE_LENGTH",
    "CMD_SET_COLOR",
    "CMD_SET_COLOR_TEMP",
    "CMD_SET_COOLING_SETPOINT",
    "CMD_SET_ENTRY_DELAY",
    "CMD_SET_EXIT_DELAY",
    "CMD_SET_FAN_MODE",
    "CMD_SET_HEATING_SETPOINT",
    "CMD_SET_HUE",
    "CMD_SET_LEVEL",
    "CMD_SET_POSITION",
    "CMD_SET_PRESENCE",
    "CMD_SET_SAT",
    "CMD_SET_SPEED",
    "CMD_SET_THERMOSTAT_MODE",
    "CMD_SIREN",
    "CMD_STROBE",
    "CMD_UNLOCK",
    "COLOR_MODE_CT",
    "COLOR_MODE_RGB",
    "ConnectionError",
    "DEFAULT_FAN_SPEEDS",
    "Device",
    "Event",
    "HSM_ARM_ALL",
    "HSM_ARM_AWAY",
    "HSM_ARM_HOME",
    "HSM_ARM_NIGHT",
    "HSM_ARM_RULES",
    "HSM_CANCEL_ALERTS",
    "HSM_DISARM",
    "HSM_DISARM_ALL",
    "HSM_DISARM_RULES",
    "HSM_STATUS_ALL_DISARMED",
    "HSM_STATUS_ARMED_AWAY",
    "HSM_STATUS_ARMED_HOME",
    "HSM_STATUS_ARMED_NIGHT",
    "HSM_STATUS_ARMING_AWAY",
    "HSM_STATUS_ARMING_HOME",
    "HSM_STATUS_ARMING_NIGHT",
    "HSM_STATUS_DISARMED",
    "Hub",
    "ID_HSM_STATUS",
    "ID_MODE",
    "InvalidConfig",
    "InvalidToken",
    "RequestError",
    "STATE_ARMED_AWAY",
    "STATE_ARMED_HOME",
    "STATE_ARMED_NIGHT",
    "STATE_CLOSED",
    "STATE_CLOSING",
    "STATE_DISARMED",
    "STATE_LOCKED",
    "STATE_LOW",
    "STATE_OFF",
    "STATE_ON",
    "STATE_OPEN",
    "STATE_OPENING",
    "STATE_PARTIALLY_OPEN",
    "STATE_UNKNOWN",
    "STATE_UNLOCKED",
    "STATE_UNLOCKED_WITH_TIMEOUT",
]



from typing import Any

from aiohttp import ClientResponse


class ConnectionError(Exception):
    """Error when hub isn't responding."""


class InvalidToken(Exception):
    """Error for invalid access token."""


class InvalidConfig(Exception):
    """Error indicating invalid hub config data."""


class InvalidMode(Exception):
    """Error indicating that a mode is invalid."""

    def __init__(self, mode: str, **kwargs):
        super().__init__(f"Invalid mode '{mode}'")


class RequestError(Exception):
    """An error indicating that a request failed."""

    def __init__(self, resp: ClientResponse, **kwargs):
        # Pyright doesn't like the @reify used on ClientResponse.url
        any_resp: Any = resp
        super().__init__(
            f"{resp.method} {any_resp.url} - [{resp.status}] {resp.reason}"
        )



from time import time
from types import MappingProxyType
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union


class Attribute:
    def __init__(self, properties: Dict[str, Any]):
        self._properties = properties

    @property
    def name(self) -> str:
        return self._properties["name"]

    @property
    def type(self) -> str:
        return self._properties["dataType"]

    @property
    def value(self) -> Union[str, float]:
        return self._properties["currentValue"]

    @property
    def values(self) -> Optional[List[str]]:
        if "values" not in self._properties:
            return None
        return self._properties["values"]

    def update_value(self, value: Union[str, float]) -> None:
        self._properties["currentValue"] = value

    def __iter__(self):
        for key in "name", "type", "value":
            yield key, getattr(self, key)

    def __str__(self):
        return f"<Attribute name={self.name} type={self.type} value={self.value}>"


class Device:
    def __init__(self, properties: Dict[str, Any]):
        self.update_state(properties)

    @property
    def id(self) -> str:
        return self._properties["id"]

    @property
    def name(self) -> str:
        return self._properties["label"]

    @property
    def type(self) -> str:
        return self._properties["name"]

    @property
    def attributes(self) -> Mapping[str, Attribute]:
        return self._attributes_ro

    @property
    def capabilities(self) -> Sequence[str]:
        return self._capabilities

    @property
    def commands(self) -> Sequence[str]:
        return self._commands

    @property
    def last_update(self) -> float:
        """
        Return the last time this device was updated as a unix timestamp.
        """
        return self._last_update

    def update_attr(self, attr_name: str, value: Union[str, int]) -> None:
        attr = self.attributes[attr_name]
        attr.update_value(value)
        self._last_update = time()

    def update_state(self, properties: Dict[str, Any]) -> None:
        self._properties = properties
        self._last_update = time()

        self._attributes: Dict[str, Attribute] = {}
        self._attributes_ro = MappingProxyType(self._attributes)
        for attr in properties.get("attributes", []):
            self._attributes[attr["name"]] = Attribute(attr)

        caps: List[str] = [
            p for p in properties.get("capabilities", []) if isinstance(p, str)
        ]
        self._capabilities: Tuple[str, ...] = tuple(caps)

        commands: List[str] = [
            p for p in properties.get("commands", []) if isinstance(p, str)
        ]
        self._commands: Tuple[str, ...] = tuple(commands)

    def __iter__(self):
        for key in "id", "name", "type", "attributes", "capabilities":
            yield key, getattr(self, key)

    def __str__(self):
        return f'<Device id="{self.id}" name="{self.name}" type="{self.type}">'


class Event:
    def __init__(self, properties: Dict[str, Any]):
        self._properties = properties

    @property
    def device_id(self) -> str:
        return self._properties["deviceId"]

    @property
    def device_name(self) -> Optional[str]:
        return self._properties.get("displayName")

    @property
    def description(self) -> Optional[str]:
        return self._properties.get("descriptionText")

    @property
    def attribute(self) -> str:
        return self._properties["name"]

    @property
    def type(self) -> Optional[str]:
        return self._properties.get("type")

    @property
    def value(self) -> Union[str, float]:
        return self._properties["value"]

    def __iter__(self):
        for key in (
            "device_id",
            "device_name",
            "attribute",
            "value",
            "description",
            "type",
        ):
            yield key, getattr(self, key)

    def __str__(self) -> str:
        return f'<Event device_id="{self.device_id}" device_name="{self.device_name}" attribute="{self.attribute}" value="{self.value}" description="{self.description}" type="{self.type}">'


class Mode:
    def __init__(self, properties: Dict[str, Any]):
        self._properties = properties

    @property
    def active(self) -> bool:
        return self._properties["active"]

    @active.setter
    def active(self, value: bool) -> None:
        self._properties["active"] = value

    @property
    def id(self) -> int:
        return self._properties.get("id", -1)

    @property
    def name(self) -> str:
        return self._properties["name"]

    def __iter__(self):
        for key in (
            "active",
            "id",
            "name",
        ):
            yield key, getattr(self, key)

    def __str__(self) -> str:
        return f'<Mode id="{self.id}" name="{self.name}" active="{self.active}"'



import asyncio
from asyncio.base_events import Server as AsyncioServer
from socket import socket as Socket
from ssl import SSLContext
import threading
from typing import Any, Callable, Dict, List, Optional, cast

from aiohttp import web

EventCallback = Callable[[Dict[str, Any]], None]


class Server:
    """A handle to a running server."""

    def __init__(
        self,
        handle_event: EventCallback,
        host: str,
        port: int,
        ssl_context: Optional[SSLContext] = None,
    ):
        """Initialize a Server."""
        self.host = host
        self.port = port
        self.handle_event = handle_event
        self.ssl_context = ssl_context
        self._main_loop = asyncio.get_event_loop()

    @property
    def url(self) -> str:
        scheme = "http" if self.ssl_context is None else "https"
        return f"{scheme}://{self.host}:{self.port}"

    def start(self) -> None:
        """Start a new server running in a background thread."""
        app = web.Application()
        app.add_routes([web.post("/", self._handle_request)])
        self._runner = web.AppRunner(app)

        self._startup_event = threading.Event()
        self._server_loop = asyncio.new_event_loop()
        t = threading.Thread(target=self._run)
        t.start()

        # Wait for server to startup
        self._startup_event.wait()

    def stop(self) -> None:
        """Gracefully stop a running server."""
        # Call the server shutdown functions and wait for them to finish. These
        # must be called on the server thread's event loop.
        future = asyncio.run_coroutine_threadsafe(self._stop(), self._server_loop)
        future.result(5)

        # Stop the server thread's event loop
        self._server_loop.call_soon_threadsafe(self._server_loop.stop)

    async def _handle_request(self, request: web.Request) -> web.Response:
        """Handle an incoming request."""
        event = await request.json()
        # This handler will be called on the server thread. Call the external
        # handler on the app thread.
        self._main_loop.call_soon_threadsafe(self.handle_event, event)
        return web.Response(text="OK")

    def _run(self) -> None:
        """Execute the server in its own thread with its own event loop."""
        asyncio.set_event_loop(self._server_loop)
        self._server_loop.run_until_complete(self._runner.setup())

        site = web.TCPSite(
            self._runner, self.host, self.port, ssl_context=self.ssl_context
        )
        self._server_loop.run_until_complete(site.start())

        # If the Server was initialized with port 0, determine what port the
        # underlying server ended up listening on
        if self.port == 0:
            site_server = cast(AsyncioServer, site._server)
            sockets = cast(List[Socket], site_server.sockets)
            socket = sockets[0]
            self.port = socket.getsockname()[1]

        self._startup_event.set()
        self._server_loop.run_forever()

    async def _stop(self) -> None:
        """Stop the server."""
        await self._runner.shutdown()
        await self._runner.cleanup()


def create_server(
    handle_event: EventCallback,
    host: str = "0.0.0.0",
    port: int = 0,
    ssl_context: Optional[SSLContext] = None,
) -> Server:
    """Create a new server."""
    return Server(handle_event, host, port, ssl_context)
